import { EditorView } from 'prosemirror-view';
import { DOMSerializer, } from 'prosemirror-model';
import { ExtensionManager } from './ExtensionManager.js';
import { EditorState } from 'prosemirror-state';
import { createNodeFromContent } from './utilities/createNodeFromContent.js';
import { CommandManager } from './commands/CommandManager.js';
export function getHTMLFromFragment(fragment, schema) {
    const documentFragment = DOMSerializer.fromSchema(schema).serializeFragment(fragment);
    const temporaryDocument = document.implementation.createHTMLDocument();
    const container = temporaryDocument.createElement('div');
    container.appendChild(documentFragment);
    return container.innerHTML;
}
function createDocument(content, schema, parseOptions = {}, options = {}) {
    return createNodeFromContent(content, schema, {
        slice: false,
        parseOptions,
        errorOnInvalidContent: options.errorOnInvalidContent,
    });
}
export class CoreEditor extends EventTarget {
    constructor(options = {}) {
        super();
        Object.defineProperty(this, "options", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: {
                element: null, // document.createElement('div'),
                extensions: [],
            }
        });
        Object.defineProperty(this, "extensionManager", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "commandManager", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "view", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "state", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.options = {
            ...this.options,
            ...options,
        };
        this.extensionManager = new ExtensionManager(this.options.extensions, this);
        // const content = this.options.content ? this.options.content : {
        //   type: this.extensionManager.schema.topNodeType.name,
        //   content: this.extensionManager.schema.topNodeType.spec.EMPTY_DOC,
        // };
        const content = this.options.content
            ? this.options.content
            : this.extensionManager.schema.topNodeType.spec.EMPTY_DOC;
        this.createView(structuredClone(content));
        // this.createView('');
        this.commandManager = new CommandManager(this, this.extensionManager.commandConstructors);
        this.setupPlugins();
    }
    get schema() {
        return this.extensionManager.schema;
    }
    chain() {
        return this.commandManager.chain();
    }
    can() {
        return this.commandManager.can();
    }
    createView(content) {
        let doc = createDocument(content, this.schema, this.options.parseOptions, { errorOnInvalidContent: false });
        this.state = EditorState.create({ doc });
        if (this.options.element) {
            this.view = new EditorView(this.options.element, {
                state: this.state,
                dispatchTransaction: (tx) => this.dispatchTransaction(tx),
            });
        }
    }
    dispatchTransaction(transaction) {
        this.state = this.state.apply(transaction);
        if (this.view) {
            this.view.updateState(this.state);
            const event = new CustomEvent('transaction', {
                detail: {
                    editor: this,
                    transaction,
                },
            });
            this.dispatchEvent(event);
        }
    }
    setupPlugins() {
        this.state = this.state.reconfigure({
            plugins: this.extensionManager.plugins,
        });
        if (this.view) {
            this.view.updateState(this.state);
            this.view.setProps({
                nodeViews: this.extensionManager.nodeViews,
            });
        }
    }
    getJSON() {
        return this.state.doc.toJSON();
    }
    /**
     * Get the document as HTML.
     */
    getHTML() {
        return getHTMLFromFragment(this.state.doc.content, this.schema);
    }
    setDocument(content, mediaType) {
        if (!content) { // clear
            content = {
                type: this.extensionManager.schema.topNodeType.name,
                content: this.extensionManager.schema.topNodeType.spec.EMPTY_DOC.content,
                // content: this.extensionManager.schema.topNodeType.createAndFill(),
            };
            mediaType = undefined;
        }
        let doc;
        if (mediaType) {
            const converter = this.extensionManager.converters[mediaType];
            if (converter) {
                doc = converter.toDoc(content);
            }
        }
        else {
            doc = createDocument(content, this.schema, this.options.parseOptions, { errorOnInvalidContent: false });
        }
        this.state = EditorState.create({
            doc,
            plugins: this.state.plugins,
            storedMarks: this.state.storedMarks,
        });
        if (this.view) {
            this.view.updateState(this.state);
        }
        const event = new CustomEvent('doc:loaded', {
            detail: {
                editor: this,
                doc,
            },
        });
        this.dispatchEvent(event);
    }
    getDocument(mediaType) {
        if (mediaType) {
            const converter = this.extensionManager.converters[mediaType];
            if (converter) {
                return converter.fromDoc(this.state.doc);
            }
        }
        return this.state.doc;
    }
}
