import { Schema } from 'prosemirror-model';
import { keymap } from 'prosemirror-keymap';
import { InputRulesPlugin, } from './plugins/input-rules/InputRulesPlugin.js';
import { chainCommands } from './commands/mod.js';
import { addAttributesToSchema } from './utilities/getHtmlAttributes.js';
export function findDuplicates(items) {
    const filtered = items.filter((el, index) => items.indexOf(el) !== index);
    return Array.from(new Set(filtered));
}
export function splitExtensions(extensions) {
    const baseExtensions = Array.from(extensions).filter((extension) => extension.type === 'extension');
    const nodeExtensions = Array.from(extensions).filter((extension) => extension.type === 'node');
    const markExtensions = Array.from(extensions).filter((extension) => extension.type === 'mark');
    return {
        baseExtensions,
        nodeExtensions,
        markExtensions,
    };
}
export class ExtensionManager {
    constructor(extensions, editor) {
        Object.defineProperty(this, "editor", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: editor
        });
        Object.defineProperty(this, "schema", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "extensions", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Set()
        });
        Object.defineProperty(this, "plugins", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "nodeViews", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: {}
        });
        Object.defineProperty(this, "commandConstructors", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: {}
        });
        Object.defineProperty(this, "converters", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: {}
        });
        Object.defineProperty(this, "debug", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: true
        });
        this.setupExtensions(extensions);
        this.schema = this.getSchemaByResolvedExtensions(editor);
        const event = new CustomEvent('schema:ready', {
            detail: {
                editor,
                schema: this.schema,
            },
        });
        editor.dispatchEvent(event);
        this.plugins = this.getPlugins();
    }
    getPlugins() {
        const plugins = [];
        const inputRules = [];
        const commands = new Map();
        const keyBindings = new Map();
        const mergeCommands = (toInsert, extName) => {
            for (const key in toInsert) {
                const commandConstructor = toInsert[key];
                if (this.debug) {
                    const wrappedConstructor = () => {
                        const realCommand = commandConstructor();
                        const command = (state, dispatch, view) => {
                            if (dispatch) {
                                console.debug(`Command: ${extName}.${key}`);
                            }
                            return realCommand(state, dispatch, view);
                        };
                        return command;
                    };
                    commands.set(key, wrappedConstructor);
                    this.commandConstructors[key] = wrappedConstructor;
                }
                else {
                    commands.set(key, commandConstructor);
                    this.commandConstructors[key] = commandConstructor;
                }
            }
        };
        function mergeShortcuts(toInsert, extName) {
            for (const key in toInsert) {
                const commandConstructor = commands.get(toInsert[key]);
                if (!commandConstructor) {
                    console.warn(`No command constructor: ${toInsert[key]}`);
                    continue;
                }
                const command = commandConstructor();
                const keyBinding = keyBindings.get(key);
                if (keyBinding) {
                    keyBindings.set(key, chainCommands(keyBinding, command));
                }
                else {
                    keyBindings.set(key, command);
                }
            }
        }
        let converters = {};
        for (const extension of this.extensions) {
            if (extension.type === 'node') {
                const nodeType = this.schema.nodes[extension.name];
                inputRules.push(...extension.getInputRules(nodeType));
                plugins.push(...extension.getProseMirrorPlugins(this.editor, this.schema));
                mergeCommands(extension.getCommands(this.editor, nodeType), extension.name);
                mergeShortcuts(extension.getKeyboardShortcuts(this.editor), extension.name);
                converters = {
                    ...converters,
                    ...extension.getConverters(this.schema),
                };
                const nodeView = extension.getNodeView();
                if (nodeView) {
                    this.nodeViews[extension.name] = nodeView;
                }
            }
            if (extension.type === 'mark') {
                const markType = this.schema.marks[extension.name];
                inputRules.push(...extension.getInputRules(markType));
                mergeCommands(extension.getCommands(this.editor, markType), extension.name);
                mergeShortcuts(extension.getKeyboardShortcuts(this.editor), extension.name);
            }
            if (extension.type === 'extension') {
                plugins.push(...extension.getProseMirrorPlugins(this.editor, this.schema));
                mergeCommands(extension.getCommands(this.editor), extension.name);
                mergeShortcuts(extension.getKeyboardShortcuts(this.editor), extension.name);
                converters = {
                    ...converters,
                    ...extension.getConverters(this.schema),
                };
            }
        }
        if (this.debug) {
            for (const key in keyBindings) {
                const wrapperCommand = (state, dispatch, view) => {
                    console.debug(`Key: ${key}`);
                    return true;
                };
                keyBindings.set(key, chainCommands(wrapperCommand, keyBindings.get(key)));
            }
        }
        this.converters = converters;
        plugins.push(new InputRulesPlugin(inputRules));
        plugins.push(keymap(Object.fromEntries(keyBindings)));
        return plugins;
    }
    setupExtensions(extensions) {
        const allExtensions = new Map();
        const createMap = (extensions) => {
            for (const extension of extensions) {
                allExtensions.set(extension.name, extension);
                if (extension.requires) {
                    const childExtensions = Array.from(extension.requires).filter((e) => typeof e !== 'string');
                    createMap(new Set(childExtensions));
                }
            }
        };
        createMap(extensions);
        const initialized = new Set();
        const initializeExtension = (extension) => {
            console.info(`Initialize ${extension.type} ${extension.name}`);
            this.extensions.add(extension);
        };
        function recursiveInitializeExtension(extension) {
            if (initialized.has(extension.name)) {
                return;
            }
            const requires = (extension.requires || []).map((e) => typeof e === 'string' ? e : e.name);
            for (const require of requires) {
                if (!initialized.has(require)) {
                    const requiredExtension = allExtensions.get(require);
                    if (!requiredExtension) {
                        throw new Error('Required extension not found: ' + require);
                    }
                    recursiveInitializeExtension(requiredExtension);
                }
            }
            initializeExtension(extension);
            initialized.add(extension.name);
            allExtensions.delete(extension.name);
        }
        for (const extension of allExtensions.values()) {
            recursiveInitializeExtension(extension);
        }
        if (allExtensions.size > 0) {
            throw new Error('Not all extensions initialized: ' +
                Array.from(allExtensions.keys()).join(', '));
        }
    }
    getSchemaByResolvedExtensions(editor) {
        const { nodeExtensions, markExtensions, baseExtensions } = splitExtensions(this.extensions);
        const nodes = {};
        for (const extension of nodeExtensions) {
            nodes[extension.name] = extension.getNodeSpec();
            addAttributesToSchema(nodes[extension.name], extension);
            if ('automerge' in extension) {
                nodes[extension.name].automerge = extension.automerge;
            }
        }
        const marks = {};
        for (const extension of markExtensions) {
            marks[extension.name] = extension.getMarkSpec();
            addAttributesToSchema(marks[extension.name], extension);
            if ('automerge' in extension) {
                marks[extension.name].automerge = extension.automerge;
            }
        }
        const spec = {
            topNode: this.editor.options.topNode || 'doc',
            nodes,
            marks,
        };
        for (const extension of baseExtensions) {
            if ('setupSpec' in baseExtensions) {
                baseExtensions.setupSpec(spec);
            }
        }
        const event = new CustomEvent('schema:spec', {
            detail: {
                editor,
                spec,
            },
        });
        editor.dispatchEvent(event);
        return new Schema(spec);
    }
}
